@@ -4,45 +4,59 @@ is linked from the engine and will be PhysicalInterface, TunnelInterface, etc. W
 the top level interface, there are sub-interface configurations that identify the
 basic settings such as ip address, network, administrative settings etc. These are
 not called directly but used as a reference to the top level interface.
+All sub interfaces are type dict. 
 """
-from collections import Sequence
+
+from smc.base.structs import NestedDict, BaseIterable
+from smc.api.exceptions import EngineCommandFailed
 
 
-class SubInterface(Sequence):
-    def __init__(self, subif):
-        self.subif = subif if subif is not None else []
+def get_sub_interface(typeof):
+    if typeof in clsmembers:
+        return clsmembers[typeof]
 
-    def __getitem__(self, i):
-        intf = self.subif[i]
-        for k, v in intf.items():
-            return SubInterface.get_subinterface(k)(v)
+    
+class SubInterfaceCollection(BaseIterable):
+    """
+    A Sub Interface collection for non-VLAN interfaces.
+    """
+    def __init__(self, interface): 
+        data = [clsmembers.get(kind)(data)
+                for intf in interface.data.get('interfaces', [])
+                for kind, data in intf.items()]
+        super(SubInterfaceCollection, self).__init__(data)
 
-    def __len__(self):
-        return len(self.subif)
-
-    @staticmethod
-    def get_subinterface(typeof):
-        for subif in [NodeInterface, SingleNodeInterface, ClusterVirtualInterface,
-                      CaptureInterface, InlineInterface]:
-            if subif.typeof == typeof:
-                return subif
+    
+class SubInterface(NestedDict):
+    def __init__(self, data):
+        super(SubInterface, self).__init__(data=data)
+    
+    def change_interface_id(self, interface_id):
+        """
+        Generic change interface ID for VLAN interfaces that are not
+        Inline Interfaces (non-VLAN sub interfaces do not have an
+        interface_id field).
+        
+        :param str, int interface_id: interface ID value
+        """
+        _, second = self.nicid.split('.')
+        self.update(nicid='{}.{}'.format(str(interface_id), second))
+    
+    def change_vlan_id(self, vlan_id):
+        """
+        Change a VLAN id
+        
+        :param str vlan_id: new vlan
+        """
+        first, _ = self.nicid.split('.')
+        self.update(nicid='{}.{}'.format(first, str(vlan_id)))
+    
+    def __getattr__(self, key):
+        return self.get(key)
 
-
-class LoopbackInterface(object):
-    @classmethod
-    def create(cls, address, network_value, **kwargs):
-        data = {'address': address,
-                'network_value': network_value,
-                'nicid': 'Loopback Interface'}
-
-        for k, v in kwargs.items():
-            data.update({k: v})
-        
-        return data
             
-class ClusterVirtualInterface(object):
+class ClusterVirtualInterface(SubInterface):
     """
-    ClusterVirtualInterface
     These interfaces (CVI) are used on cluster devices and applied to layer 3
     interfaces. They specify a 'VIP' (or shared IP) to be used for traffic load
     balancing or high availability. Each engine will still also have a 'node'
@@ -60,8 +74,8 @@ class ClusterVirtualInterface(object):
     typeof = 'cluster_virtual_interface'
 
     def __init__(self, data):
-        self.__dict__ = data
-
+        super(ClusterVirtualInterface, self).__init__(data)
+            
     @classmethod
     def create(cls, interface_id, address, network_value,
                **kwargs):
@@ -69,7 +83,7 @@ class ClusterVirtualInterface(object):
         :param int interface_id: nic id to use for physical interface
         :param str address: address of CVI
         :param str network_value: network for CVI
-        :rtype: dict
+        :rtype: ClusterVirtualInterface
         """
         data = {'address': address,
                 'network_value': network_value,
@@ -81,10 +95,6 @@ class ClusterVirtualInterface(object):
 
         return cls(data)
 
-    @property
-    def data(self):
-        return {'cluster_virtual_interface': self.__dict__}
-
     @property
     def vlan_id(self):
         """
@@ -98,9 +108,6 @@ class ClusterVirtualInterface(object):
             v = nicid.split('.')
             if len(v) > 1:
                 return nicid.split('.')[1]
-    
-    def __getattr__(self, attr):
-        return None
 
     def __repr__(self):
         if self.vlan_id:
@@ -111,7 +118,7 @@ class ClusterVirtualInterface(object):
                 self.__class__.__name__, self.address)
 
 
-class InlineInterface(object):
+class InlineInterface(SubInterface):
     """
     InlineInterface
     This interface type is used on layer 2 or IPS related engines. It requires
@@ -124,14 +131,14 @@ class InlineInterface(object):
     :ivar boolean inspect_unspecified_vlans: promiscuous SPAN on unspecified VLANs
     :ivar str logical_interface_ref (required): logical interface to use, by href
     :ivar str failure_mode: normal or bypass
-    :ivar str nicid: interfaces for inline pair, for example, '1-2', '5-6' (interfaces 5 and 6)
+    :ivar str nicid: interfaces for inline pair, for example, '4.50-5.55', '5-6'
     :ivar str vlan_id: vlan identifier for interface
     :ivar str zone_ref (optional): zone for second interface in pair
     """
     typeof = 'inline_interface'
 
     def __init__(self, data):
-        self.__dict__ = data
+        super(InlineInterface, self).__init__(data)
 
     @classmethod
     def create(cls, interface_id, logical_interface_ref,
@@ -142,8 +149,7 @@ class InlineInterface(object):
         :param str zone_ref: reference to zone, set on second inline pair
         :rtype: dict
         """
-        data = {'failure_mode': 'normal',
-                'inspect_unspecified_vlans': True,
+        data = {'inspect_unspecified_vlans': True,
                 'nicid': interface_id,
                 'logical_interface_ref': logical_interface_ref,
                 'zone_ref': zone_ref}
@@ -152,11 +158,49 @@ class InlineInterface(object):
             data.update({k: v})
 
         return cls(data)
+    
+    def change_vlan_id(self, vlan_id):
+        """
+        Change a VLAN id for an inline interface.
         
-    @property
-    def data(self):
-        return {'inline_interface': self.__dict__}
+        :param str vlan_id: New VLAN id. Can be in format '1-2' or
+            a single numerical value. If in '1-2' format, this specifies
+            the vlan ID for the first inline interface and the rightmost
+            for the second.
+        :return: None
+        """
+        first, second = self.nicid.split('-')
+        firstintf = first.split('.')[0]
+        secondintf = second.split('.')[0]
+        newvlan = str(vlan_id).split('-')
+        self.update(nicid='{}.{}-{}.{}'.format(
+            firstintf, newvlan[0], secondintf, newvlan[-1]))
 
+    def change_interface_id(self, newid):
+        """
+        Change the inline interface ID. The current format is
+        nicid='1-2', where '1' is the top level interface ID (first),
+        and '2' is the second interface in the pair. Consider the existing
+        nicid in case this is a VLAN.
+        
+        :param str newid: string defining new pair, i.e. '3-4'
+        :return: None
+        """
+        try:
+            newleft, newright = newid.split('-')
+        except ValueError:
+            raise EngineCommandFailed('You must provide two parts when changing '
+                'the interface ID on an inline interface, i.e. 1-2.')
+        first, second = self.nicid.split('-')
+        if '.' in first and '.' in second:
+            firstvlan = first.split('.')[-1]
+            secondvlan = second.split('.')[-1]
+            self.update(nicid='{}.{}-{}.{}'.format(
+                newleft, firstvlan, newright, secondvlan))
+        else:
+            # Top level interface or no VLANs
+            self.update(nicid=newid)
+    
     @property
     def vlan_id(self):
         """
@@ -173,15 +217,44 @@ class InlineInterface(object):
                     u.append(vlan.split('.')[-1])
             return '-'.join(u)
 
-    def __getattr__(self, attr):
-        return None
-
     def __repr__(self):
         return '{0}(nicid={1})'.format(
             self.__class__.__name__, self.nicid)
 
 
-class CaptureInterface(object):
+class InlineIPSInterface(InlineInterface):
+    """
+    An Inline IPS Interface is a new interface type introduced
+    in SMC version 6.3. This interface type is the same as a
+    normal IPS interface except that it is applied on a Layer 3
+    Firewall.
+    
+    .. versionadded:: 0.5.6
+        Requires SMC 6.3.
+    """
+    typeof = 'inline_ips_interface'
+    
+    def __init__(self, data):
+        super(InlineIPSInterface, self).__init__(data)
+    
+
+class InlineL2FWInterface(InlineInterface):
+    """
+    An Inline L2FW Interface is a new interface type introduced
+    in SMC version 6.3. This interface type is the a layer 2 FW
+    interface on a layer 3 firewall. By default this interface
+    type does not support bypass mode and will discard on overload.
+    
+    .. versionadded:: 0.5.6
+        Requires SMC 6.3.
+    """
+    typeof = 'inline_l2fw_interface'
+    
+    def __init__(self, data):
+        super(InlineL2FWInterface, self).__init__(data)
+        
+
+class CaptureInterface(SubInterface):
     """
     Capture Interface (SPAN)
     This is a single physical interface type that can be installed on either
@@ -196,8 +269,8 @@ class CaptureInterface(object):
     typeof = 'capture_interface'
 
     def __init__(self, data):
-        self.__dict__ = data
-
+        super(CaptureInterface, self).__init__(data)
+        
     @classmethod
     def create(cls, interface_id, logical_interface_ref, **kwargs):
         """
@@ -216,18 +289,16 @@ class CaptureInterface(object):
         return cls(data)
 
     @property
-    def data(self):
-        return {'capture_interface': self.__dict__}
-
-    def __getattr__(self, attr):
-        return None
+    def vlan_id(self):
+        if '.' in self.nicid:
+            return self.nicid.split('.')[-1]
 
     def __repr__(self):
         return '{0}(nicid={1})'.format(
             self.__class__.__name__, self.nicid)
 
 
-class NodeInterface(object):
+class NodeInterface(SubInterface):
     """
     Node Interface
     Node dedicated interface (NDI) is used on specific engine types and represents
@@ -276,7 +347,7 @@ class NodeInterface(object):
     typeof = 'node_interface'
 
     def __init__(self, data):
-        self.__dict__ = data
+        super(NodeInterface, self).__init__(data)
 
     @classmethod
     def create(cls, interface_id, address, network_value,
@@ -302,11 +373,7 @@ class NodeInterface(object):
             data.update({k: v})
 
         return cls(data)
-
-    @property
-    def data(self):
-        return {self.typeof: self.__dict__}
-
+    
     @property
     def vlan_id(self):
         """
@@ -321,9 +388,6 @@ class NodeInterface(object):
             if len(v) > 1:
                 return nicid.split('.')[1]
 
-    def __getattr__(self, attr):
-        return None
-
     def __repr__(self):
         if self.vlan_id:
             return '{0}(address={1}, vlan_id={2})'.format(
@@ -346,7 +410,6 @@ class SingleNodeInterface(NodeInterface):
 
     def __init__(self, data):
         super(SingleNodeInterface, self).__init__(data)
-        pass
 
     @classmethod
     def create(cls, interface_id, address, network_value,
@@ -406,6 +469,170 @@ class SingleNodeInterface(NodeInterface):
         return cls(data)
 
 
+class LoopbackClusterInterface(ClusterVirtualInterface):
+    """
+    This represents the CVI Loopback IP address.
+    A CVI loopback IP address is used for loopback traffic that is sent to
+    the whole cluster. It is shared by all the nodes in the cluster.
+    """
+    typeof = 'loopback_cluster_virtual_interface'
+    
+    def __init__(self, data, engine=None):
+        self._engine = engine
+        super(LoopbackClusterInterface, self).__init__(data)
+        
+    @classmethod
+    def create(cls, address, ospf_area=None, **kwargs):
+        """
+        Create a loopback interface. Uses parent constructor
+        
+        :rtype: LoopbackClusterInterface
+        """
+        return super(LoopbackClusterInterface, cls).create(
+            address=address,
+            network_value='{}/32'.format(address),
+            interface_id='Loopback Interface',
+            ospfv2_area_ref=ospf_area,
+            **kwargs)
+    
+    def delete(self):
+        """
+        Delete a loopback cluster virtual interface from this engine. 
+        Changes to the engine configuration are done immediately.
+        
+        You can find cluster virtual loopbacks by iterating at the
+        engine level::
+        
+            for loopbacks in engine.loopback_interface:
+                ...
+        
+        :raises UpdateElementFailed: failure to delete loopback interface
+        :return: None
+        """
+        self._engine.data[self.typeof] = \
+            [loopback for loopback in self._engine.data.get(self.typeof, [])
+             if loopback.get('address') != self.address]
+            
+        self._engine.update()
+    
+    def add_node_loopback(self, nodes, ospf_area=None):
+        """
+        Add loopback interfaces to a cluster. When adding a loopback on a
+        cluster, every cluster node must have a loopback defined or you
+        can optionally configure a loopback CVI address.
+        
+        Nodes should be in the format::
+        
+            {'address': '127.0.0.10', 'nodeid': 1,
+             'address': '127.0.0.11', 'nodeid': 2}
+             
+        :param dict nodes: nodes defintion for cluster nodes
+        :param str ospf_area: optional OSPF area for this loopback
+        :raises EngineCommandFailed: failed creating loopback
+        """
+        pass
+    
+    def add_cvi_loopback(self, address, ospf_area=None, **kw):
+        """
+        Add a loopback interface as a cluster virtual loopback. This enables
+        the loopback to 'float' between cluster members. Changes are committed
+        immediately.
+        
+        :param str address: ip address for loopback
+        :param int rank: rank of this entry
+        :param str,Element ospf_area: optional ospf_area to add to loopback
+        :raises UpdateElementFailed: failure to save loopback address
+        :return: None
+        """
+        lb = self.create(address, ospf_area, **kw)
+       
+        if self.typeof in self._engine.data:
+            self._engine.data[self.typeof].append(lb.data)
+        else:
+            self._engine.data[self.typeof] = [lb.data]
+        
+        self._engine.update()
+  
+    def __repr__(self):
+        return 'LoopbackClusterInterface(address={}, auth_request={})'.format(
+            self.address, self.auth_request)
+        
+
+class LoopbackInterface(NodeInterface):
+    """
+    Loopback interface for a physical or virtual single firewall.
+    To create a loopback interface, call from the engine node::
+    
+        engine.loopback_interface.add_single(...)
+    """
+    typeof = 'loopback_node_dedicated_interface'
+    
+    def __init__(self, data, engine=None):
+        self._engine = engine
+        super(LoopbackInterface, self).__init__(data)
+        
+    @classmethod
+    def create(cls, address, rank=1, nodeid=1, ospf_area=None, **kwargs):
+        return super(LoopbackInterface, cls).create(
+            interface_id='Loopback Interface',
+            #rank=rank,
+            address=address,
+            network_value='{}/32'.format(address),
+            nodeid=nodeid,
+            ospfv2_area_ref=ospf_area,
+            **kwargs)
+    
+    def add_single(self, address, rank=1, nodeid=1, ospf_area=None, **kwargs):
+        """
+        Add a single loopback interface to this engine. This is used
+        for single or virtual FW engines.
+        
+        :param str address: ip address for loopback
+        :param int nodeid: nodeid to apply. Default to 1 for single FW
+        :param str, Element ospf_area: ospf area href or element
+        :raises UpdateElementFailed: failure to create loopback address
+        :return: None
+        """
+        lb = self.create(address, rank, nodeid, ospf_area, **kwargs)
+        self._engine.nodes[0].data[self.typeof].append(lb.data)
+        self._engine.update()
+    
+    def delete(self):
+        """
+        Delete a loopback interface from this engine. Changes to the
+        engine configuration are done immediately.
+        
+        A simple way to obtain an existing loopback is to iterate the
+        loopbacks or to get by address::
+        
+            lb = engine.loopback_interface.get('127.0.0.10')
+            lb.delete()
+        
+        .. warning:: When deleting a loopback assigned to a node on a cluster
+            all loopbacks with the same rank will also be removed.
+        
+        :raises UpdateElementFailed: failure to delete loopback interface
+        :return: None
+        """
+        nodes = []
+        for node in self._engine.nodes:
+            node.data[self.typeof] = \
+                [lb for lb in node.loopback_node_dedicated_interface
+                 if lb.get('rank') != self.rank]
+            nodes.append({node.type: node.data})
+        
+        self._engine.data['nodes'] = nodes
+        self._engine.update()
+    
+    def change_ipaddress(self, address):
+        self.update(address=address,
+                    network_value='{}/32'.format(address))
+
+    def __repr__(self):
+        return 'LoopbackInterface(address={}, nodeid={}, rank={})'.format(
+            self.address, self.nodeid, self.rank)   
+
+
 def _add_vlan_to_inline(inline_intf, vlan_id, vlan_id2=None):
     """
     Modify InlineInterface to add VLAN by ID
@@ -428,3 +655,17 @@ def _add_vlan_to_inline(inline_intf, vlan_id, vlan_id2=None):
             pass
     vals.update(nicid=nicid)
     return inline_intf
+
+
+def inheritors(klass):
+    subclasses = {}
+    work = [klass]
+    while work:
+        parent = work.pop()
+        for child in parent.__subclasses__():
+            if child.typeof not in subclasses:
+                subclasses[child.typeof] = child
+                work.append(child)
+    return subclasses
+
+clsmembers = inheritors(SubInterface)
