@@ -20,10 +20,10 @@ overidden.
           needed, calling open() will lock the policy from test_external modifications
           until save() is called.
 """
-from smc.api.exceptions import TaskRunFailed, PolicyCommandFailed,\
-    ResourceNotFound
-from smc.administration.tasks import TaskOperationPoller
-from smc.base.model import Element, prepared_request, lookup_class
+import collections
+from smc.api.exceptions import PolicyCommandFailed
+from smc.administration.tasks import Task
+from smc.base.model import Element, lookup_class
 
 
 class Policy(Element):
@@ -38,10 +38,6 @@ class Policy(Element):
     'export', and 'upload' are encapsulated into this base class.
     """
 
-    def __init__(self, name, **meta):
-        super(Policy, self).__init__(name, **meta)
-        pass
-
     def upload(self, engine, timeout=5, wait_for_finish=False, **kw):
         """
         Upload policy to specific device. Using wait for finish
@@ -58,17 +54,8 @@ class Policy(Element):
         :raises: TaskRunFailed
         :return: TaskOperationPoller
         """
-        task = prepared_request(
-            TaskRunFailed,
-            href=self.data.get_link('upload'),
-            params={'filter': engine}
-            ).create().json
-
-        return TaskOperationPoller(
-            task=task,
-            timeout=timeout,
-            wait_for_finish=wait_for_finish,
-            **kw)
+        return Task.execute(self, 'upload', params={'filter': engine},
+            timeout=timeout, wait_for_finish=wait_for_finish, **kw)
 
     def open(self):
         """
@@ -80,13 +67,10 @@ class Policy(Element):
         :raises PolicyCommandFailed: Cannot open policy
         :return: None
         """
-        try:
-            prepared_request(
-                PolicyCommandFailed,
-                href=self.data.get_link('open')
-            ).create()
-        except ResourceNotFound:
-            pass
+        self.make_request(
+            PolicyCommandFailed,
+            method='create',
+            resource='open')
 
     def save(self):
         """ Save policy that was modified
@@ -94,23 +78,21 @@ class Policy(Element):
 
         :return: None
         """
-        try:
-            prepared_request(
-                PolicyCommandFailed,
-                href=self.data.get_link('save')
-            ).create()
-        except ResourceNotFound:
-            pass
+        self.make_request(
+            PolicyCommandFailed,
+            method='create',
+            resource='save')
 
     def force_unlock(self):
-        """ Forcibly unlock a locked policy
+        """
+        Forcibly unlock a locked policy
 
         :return: None
         """
-        prepared_request(
+        self.make_request(
             PolicyCommandFailed,
-            href=self.data.get_link('force_unlock')
-        ).create()
+            method='create',
+            resource='force_unlock')
 
     def search_rule(self, search):
         """
@@ -126,16 +108,17 @@ class Policy(Element):
         :return: rule elements matching criteria
         :rtype: list(Element)
         """
-        result = prepared_request(
-            href=self.data.get_link('search_rule'),
-            params={'filter': search}
-        ).read()
-        if result.json:
+        result = self.make_request(
+            resource='search_rule',
+            params={'filter': search})
+        
+        if result:
             results = []
-            for data in result.json:
-                if data.get('type') == 'ips_ethernet_rule':
+            for data in result:
+                if 'ethernet' in data.get('type'):
                     klazz = lookup_class('ethernet_rule')
-                elif data.get('type') == 'ips_ipv4_access_rule':
+                elif data.get('type') in [
+                    'ips_ipv4_access_rule', 'l2_interface_ipv4_access_rule']:
                     klazz = lookup_class('layer2_ipv4_access_rule')
                 else:
                     klazz = lookup_class(data.get('type'))
@@ -143,9 +126,6 @@ class Policy(Element):
             return results
         return []
 
-    def search_category_tags_from_element(self):
-        pass
-
     @property
     def template(self):
         """
@@ -167,6 +147,46 @@ class Policy(Element):
         """
         return Element.from_href(self.data.get('inspection_policy'))
 
+    @property
+    def file_filtering_policy(self):
+        """
+        Each policy is required to have a reference to a File Filtering
+        Policy. To use you will need a rule in the policy that has an
+        action with file filtering turned on.
+        
+        :return :class:`smc.policy.file_filtering.FileFilteringPolicy`
+        """
+        if 'file_filtering_policy' in self.data: # Workaround for SMC 6.3
+            return Element.from_href(self.data['file_filtering_policy'])
+    
+    def rule_counters(self, engine, duration_type='one_week',
+            duration=0, start_time=0):
+        """
+        .. versionadded:: 0.5.6
+            Obtain rule counters for this policy. Requires SMC >= 6.2
+        
+        Rule counters can be obtained for a given policy and duration for
+        those counters can be provided in duration_type. A custom start
+        range can also be provided.
+
+        :param Engine engine: the target engine to obtain rule counters from
+        :param str duration_type: duration for obtaining rule counters. Valid
+            options are: one_day, one_week, one_month, six_months, one_year,
+            custom, since_last_upload
+        :param int duration: if custom set for duration type, specify the
+            duration in seconds (Default: 0)
+        :param int start_time: start time in milliseconds (Default: 0)
+        :raises: ActionCommandFailed
+        :return: list of rule counter objects
+        :rtype: RuleCounter
+        """
+        json = {'target_ref': engine.href, 'duration_type': duration_type}
+        return [RuleCounter(**rule)
+                for rule in self.make_request(
+                    method='create',
+                    resource='rule_counter',
+                    json=json)]
+
 
 class InspectionPolicy(Policy):
     """
@@ -178,10 +198,26 @@ class InspectionPolicy(Policy):
     """
     typeof = 'inspection_template_policy'
 
-    def __init__(self, name, **meta):
-        super(InspectionPolicy, self).__init__(name, **meta)
-        pass
-
     def export(self): pass  # Not valid for inspection policy
 
     def upload(self): pass  # Not valid for inspection policy
+    
+
+class RuleCounter(collections.namedtuple(
+        'RuleCounter', 'hits rule_ref total_hits')):
+    """
+    Rule counter representing hits for a specific rule.
+    
+    :param int hits: hits for this given rule
+    :param rule_ref: rule reference to obtain the rule
+    :param total_hits: total number of hits over the duration
+    """
+    __slots__ = ()
+    
+    def __new__(cls, rule_ref, hits=0, total_hits=0):  # @ReservedAssignment
+        return super(RuleCounter, cls).__new__(cls, hits, rule_ref, total_hits)
+    
+    @property
+    def rule(self):
+        return Element.from_href(self.rule_ref)
+
