@@ -1,18 +1,19 @@
-from smc.base.model import Element, ElementCreator, prepared_request, SubElement
+from smc.base.model import Element, ElementCreator, SubElement
 from smc.api.exceptions import CreatePolicyFailed, CreateElementFailed,\
     PolicyCommandFailed, ElementNotFound
 from smc.base.collection import sub_collection
-from smc.base.util import element_resolver
+from smc.vpn.elements import VPNProfile, VPNSite
+from smc.base.decorators import cached_property
 
 
-class VPNPolicy(Element):
+class PolicyVPN(Element):
     """
     Create a new VPN Policy.
     ::
     
-        >>> VPNPolicy.create(name='myvpn')
-        VPNPolicy(name=myvpn)
-        >>> v = VPNPolicy('myvpn')
+        >>> PolicyVPN.create(name='myvpn')
+        PolicyVPN(name=myvpn)
+        >>> v = PolicyVPN('myvpn')
         >>> print(v.vpn_profile)
         VPNProfile(name=VPN-A Suite)
 
@@ -22,10 +23,6 @@ class VPNPolicy(Element):
     """
     typeof = 'vpn'
 
-    def __init__(self, name, **meta):
-        super(VPNPolicy, self).__init__(name, **meta)
-        pass
-
     @classmethod
     def create(cls, name, nat=False, mobile_vpn_toplogy_mode=None,
                vpn_profile=None):
@@ -35,21 +32,20 @@ class VPNPolicy(Element):
         :param name: name of vpn policy
         :param bool nat: whether to apply NAT to the VPN (default False)
         :param mobile_vpn_toplogy_mode: whether to allow remote vpn
-        :param str vpn_profile: reference to VPN profile, or uses default
-        :rtype: VPNPolicy
+        :param VPNProfile vpn_profile: reference to VPN profile, or uses default
+        :rtype: PolicyVPN
         """
-        if vpn_profile is not None:
-            vpn_profile = element_resolver(vpn_profile)
-        json = {'mobile_vpn_topology_mode': None,
+        vpn_profile = vpn_profile if vpn_profile else VPNProfile('VPN-A Suite')
+        
+        json = {'mobile_vpn_topology_mode': mobile_vpn_toplogy_mode,
                 'name': name,
                 'nat': nat,
-                'vpn_profile': vpn_profile}
+                'vpn_profile': vpn_profile.href}
 
         try:
             return ElementCreator(cls, json)
         except CreateElementFailed as err:
-            raise CreatePolicyFailed('VPN Policy create failed. Reason: {}'
-                                     .format(err))
+            raise CreatePolicyFailed(err)
 
     @property
     def nat(self):
@@ -73,7 +69,7 @@ class VPNPolicy(Element):
         else:
             self.data['nat'] = True
 
-    @property
+    @cached_property
     def vpn_profile(self):
         """
         Specified VPN Profile used by this VPN Policy
@@ -87,23 +83,21 @@ class VPNPolicy(Element):
         """
         Central Gateway Node acts as the hub of a hub-spoke VPN. 
 
-        :return: collection of :class:`GatewayNode`
-        :rtype: SubElementCollection
+        :rtype: SubElementCollection(GatewayNode)
         """
         return sub_collection(
-            self.data.get_link('central_gateway_node'),
+            self.get_relation('central_gateway_node'),
             type('CentralGatewayNode', (GatewayNode,), {}))
-
+        
     @property
     def satellite_gateway_node(self):
         """
         Node level settings for configured satellite gateways
 
-        :return: collection of :class:`GatewayNode`
-        :rtype: SubElementCollection
+        :rtype: SubElementCollection(GatewayNode)
         """
         return sub_collection(
-            self.data.get_link('satellite_gateway_node'),
+            self.get_relation('satellite_gateway_node'),
             type('SatelliteGatewayNode', (GatewayNode,), {}))
 
     @property
@@ -112,48 +106,71 @@ class VPNPolicy(Element):
         Mobile Gateway's are represented by client endpoints connecting
         to the policy based VPN.
 
-        :return: collection of :class:`GatewayNode`
-        :rtype: SubElementCollection
+        :rtype: SubElementCollection(GatewayNode)
         """
         return sub_collection(
-            self.data.get_link('mobile_gateway_node'),
+            self.get_relation('mobile_gateway_node'),
             type('MobileGatewayNode', (GatewayNode,), {}))
 
+    @property
+    def tunnels(self):
+        """
+        Return all tunnels for this VPN. A tunnel is defined as two end
+        points within the VPN topology. Endpoints are automatically
+        configureed based on whether they are a central gateway or 
+        satellite gateway. This provides access to enabling/disabling
+        and setting the preshared key for the linked endpoints.
+        List all tunnel mappings for this policy vpn::
+        
+            for tunnel in policy.tunnels:    
+                tunnela = tunnel.tunnel_side_a
+                tunnelb = tunnel.tunnel_side_b
+                print(tunnela.gateway)
+                print(tunnelb.gateway)
+    
+        :rtype: SubElementCollection(GatewayTunnel)
+        """
+        return sub_collection(
+            self.get_relation('gateway_tunnel'), GatewayTunnel)
+
     def open(self):
         """
-        Open the policy for editing
+        Open the policy for editing. This is only a valid method for
+        SMC version <= 6.1
 
-        :return: None
         :raises PolicyCommandFailed: couldn't open policy with reason
+        :return: None
         """
-        prepared_request(
+        self.make_request(
             PolicyCommandFailed,
-            href=self.data.get_link('open')
-        ).create()
+            method='create',
+            resource='open')
 
     def save(self):
         """
-        Save the policy after editing
+        Save the policy after editing. This is only a valid method for
+        SMC version <= 6.1
 
-        :return: None
         :raises PolicyCommandFailed: save failed with reason
+        :return: None
         """
-        prepared_request(
+        self.make_request(
             PolicyCommandFailed,
-            href=self.data.get_link('save')
-        ).create()
+            method='create',
+            resource='save')
 
     def close(self):
         """
-        Close the policy 
+        Close the policy. This is only a valid method for
+        SMC version <= 6.1
 
         :raises PolicyCommandFailed: close failed with reason
         :return: None
         """
-        prepared_request(
+        self.make_request(
             PolicyCommandFailed,
-            href=self.data.get_link('close')
-        ).create()
+            method='create',
+            resource='close')
 
     def validate(self):
         """
@@ -163,12 +180,8 @@ class VPNPolicy(Element):
         :return: status as string
         :rtype: str
         """
-        return self.data.get_json('validate').get('value')
-
-    def gateway_tunnel(self):
-        """
-        """
-        return self.data.get_json('gateway_tunnel')
+        return self.make_request(
+            resource='validate').get('value')
 
     def add_central_gateway(self, gateway):
         """ 
@@ -177,16 +190,16 @@ class VPNPolicy(Element):
         :param gateway: href for internal gateway or test_external gateway.
                If this is another SMC managed gateway, you can retrieve the 
                href after loading the engine. 
-               See :py:class:`smc.core.engines.Engine.internal_gateway`
+               See :py:class:`smc.core.engines.Engine.vpn`
         :raises PolicyCommandFailed: could not add gateway
         :return: None
         """
-        prepared_request(
+        self.make_request(
             PolicyCommandFailed,
-            href=self.data.get_link('central_gateway_node'),
+            method='create',
+            resource='central_gateway_node',
             json={'gateway': gateway,
-                  'node_usage': 'central'}
-        ).create()
+                  'node_usage': 'central'})
 
     def add_satellite_gateway(self, gateway):
         """
@@ -198,16 +211,16 @@ class VPNPolicy(Element):
         :param gateway: href for internal gateway or test_external gateway.
                If this is another SMC managed gateway, you can retrieve the 
                href after loading the engine. 
-               See :py:class:`smc.core.engines.Engine.internal_gateway` 
+               See :py:class:`smc.core.engines.Engine.vpn` 
         :raises PolicyCommandFailed: could not add gateway
         :return: None
         """
-        prepared_request(
+        self.make_request(
             PolicyCommandFailed,
-            href=self.data.get_link('satellite_gateway_node'),
+            method='create',
+            resource='satellite_gateway_node',
             json={'gateway': gateway,
-                  'node_usage': 'satellite'}
-        ).create()
+                  'node_usage': 'satellite'})
 
     @staticmethod
     def add_internal_gateway_to_vpn(internal_gateway_href, vpn_policy,
@@ -223,7 +236,7 @@ class VPNPolicy(Element):
         :rtype: bool
         """
         try:
-            vpn = VPNPolicy(vpn_policy)
+            vpn = PolicyVPN(vpn_policy)
             vpn.open()
             if vpn_role == 'central':
                 vpn.add_central_gateway(internal_gateway_href)
@@ -243,7 +256,7 @@ class GatewayNode(SubElement):
     This template class will return these as a collection. Gateway Node
     references need to be obtained from a VPN Policy reference::
 
-        >>> vpn = VPNPolicy('sg_vm_vpn')
+        >>> vpn = PolicyVPN('sg_vm_vpn')
         >>> vpn.open()
         >>> for gw in vpn.central_gateway_node.all():
         ...   list(gw.enabled_sites)
@@ -252,16 +265,23 @@ class GatewayNode(SubElement):
         >>> vpn.close()
     """
 
-    def __init__(self, **meta):
-        super(GatewayNode, self).__init__(**meta)
-        pass
-
+    @cached_property
+    def gateway(self):
+        """
+        The VPN gateway for this node. This is either an internal gateway
+        or an external gateway
+        
+        :return: the VPN gateway
+        :rtype: Element
+        """ 
+        return Element.from_href(self.data['gateway'])
+        
     @property
     def name(self):
         """
         Get the name from the gateway_profile reference
         """
-        return Element.from_href(self.data['gateway']).name
+        return self.gateway.name
     
     @property
     def enabled_sites(self):
@@ -269,11 +289,10 @@ class GatewayNode(SubElement):
         Return a collection of VPN Site elements that are enabled
         for this VPN gateway.
 
-        :return: collection of :class:`smc.vpn.elements.VPNSite`
-        :rtype: SubElementCollection
+        :rtype: SubElementCollection(VPNSite)
         """
         return sub_collection(
-            self.data.get_link('enabled_vpn_site'), GatewayTreeNode)
+            self.get_relation('enabled_vpn_site'), GatewayTreeNode)
 
     @property
     def disabled_sites(self):
@@ -281,11 +300,11 @@ class GatewayNode(SubElement):
         Return a collection of VPN Site elements that are disabled
         for this VPN gateway.
 
-        :return: collection of :class:`smc.vpn.elements.VPNSite`
-        :rtype: SubElementCollection
+        :rtype: SubElementCollection(VPNSite)
         """
         return sub_collection(
-            self.data.get_link('disabled_vpn_site'), GatewayTreeNode)
+            self.get_relation('disabled_vpn_site'), GatewayTreeNode)
+
 
 class GatewayTreeNode(SubElement):
     """
@@ -313,10 +332,10 @@ class GatewayTreeNode(SubElement):
         :raises PolicyCommandFailed: enabling or disabling failed
         :return: None
         """
-        prepared_request(
+        self.make_request(
             PolicyCommandFailed,
-            href=self.data.get_link('self')
-            ).delete()
+            method='delete',
+            resource='self')
     
     @property
     def vpn_site(self):
@@ -326,7 +345,7 @@ class GatewayTreeNode(SubElement):
         :return VPNSite element
         :rtype: VPNSite
         """
-        return Element.from_href(self.data.get('vpn_site'))
+        return VPNSite(href=self.data.get('vpn_site'))
     
     def __str__(self):
         return '{0}(name={1})'.format(
@@ -334,4 +353,83 @@ class GatewayTreeNode(SubElement):
 
     def __repr__(self):
         return str(self)
-            
\ No newline at end of file
+    
+
+class GatewayTunnel(SubElement):
+    """
+    A gateway tunnel represents the point to point connection
+    between two IPSEC endpoints in a PolicyVPN configuration. 
+    The tunnel arrangement is based on whether the nodes are placed
+    as a central gateway or a satellite gateway. This provides access
+    to see the point to point connections, whether the link is enabled,
+    and setting the presharred key.
+    
+    .. note:: Setting the preshared key is only required if using an
+        ExternalGateway element as one side of the VPN. Preshared keys
+        are generated automatically but read only, therefore if two
+        gateways are internally managed by SMC, the key is generated and
+        shared between the gateways automatically. However for external
+        gateways, you must set a new key to provide the same value to
+        the remote gateway.
+    """
+
+    def enable_disable(self):
+        """
+        Enable or disable the tunnel link between endpoints.
+        
+        :raises UpdateElementFailed: failed with reason
+        :return: None
+        """
+        if self.enabled:
+            self.update(enabled=False)
+        else:
+            self.update(enabled=True)
+    
+    @property
+    def enabled(self):
+        """          
+        Whether the VPN link between endpoints is enabled
+        
+        :rtype: bool
+        """
+        return self.data.get('enabled', False)
+    
+    def preshared_key(self, key):
+        """
+        Set a new preshared key for the IPSEC endpoints.
+        
+        :param str key: shared secret key to use
+        :raises UpdateElementFailed: fail with reason
+        :return: None
+        """
+        self.update(preshared_key=key)
+    
+    @property
+    def tunnel_side_a(self):
+        """
+        Return the gateway node for tunnel side A. This will
+        be an instance of GatewayNode.
+        
+        :rtype: GatewayNode
+        """
+        return type('TunnelSideA', (GatewayNode,), {
+            'href': self.data.get('gateway_node_1')})()
+    
+    @property
+    def tunnel_side_b(self):
+        """
+        Return the gateway node for tunnel side B. This will
+        be an instance of GatewayNode.
+        
+        :rtype: GatewayNode
+        """
+        return type('TunnelSideB', (GatewayNode,), {
+            'href': self.data.get('gateway_node_2')})()
+    
+    def __str__(self):
+        return '{0}(tunnel_side_a={1},tunnel_side_b={2})'.format(
+            self.__class__.__name__, self.tunnel_side_a.name, self.tunnel_side_b.name)
+
+    def __repr__(self):
+        return str(self)
+        
\ No newline at end of file
